""" Data interface for a tree of key:value pairs.

e.g., Any level of nesting of dict and list objects.
"""

from __future__ import annotations
from warnings import warn
from typing import Any
from pyqt_ext.tree import AbstractTreeItem


class KeyValueTreeItem(AbstractTreeItem):
    
    def __init__(self, data: dict | list | str | None, parent: KeyValueTreeItem | None = None) -> None:
        # A key:value mapping (dict or list) if this is the root item.
        # Otherwise a key into the parent item's key:value mapping (ignored if parent is a list).
        self._data: dict | list | str | None = data

        AbstractTreeItem.__init__(self, parent=parent)
        self.setupSubtree()
    
    def key(self) -> str | int | None:
        if isinstance(self._data, dict) or isinstance(self._data, list):
            # no key for the root key:value map
            return
        
        # this item stores a key into its parent key:value map.
        parent: KeyValueTreeItem = self.parent()
        if parent is None:
            # default to the stored key
            return self._data or self.siblingIndex()
        parent_map: dict | list = parent.value()
        if isinstance(parent_map, dict):
            # this item stores a key into the parent dict
            if (self._data is None) or (self._data == ''):
                # ensure a valid key
                key = self.uniqueName('?', list(parent_map.keys()))
                parent_map[key] = None
                self._data = key
            return self._data
        elif isinstance(parent_map, list):
            # the list index is taken from the order of the parent item's children
            return self.siblingIndex()
    
    def setKey(self, key: str) -> None:
        if isinstance(self._data, dict) or isinstance(self._data, list):
            # no key for the root key:value map
            return
        
        # this item stores a key into its parent key:value map
        parent: KeyValueTreeItem = self.parent()
        if parent is None:
            return
        parent_map: dict | list = parent.value()
        if isinstance(parent_map, dict):
            # must have a valid key
            if key is None:
                key = 'None'
            elif key == '':
                return
            elif type(key) not in [str, int, float, bool]:
                return
            # check if key has changed
            if key == self.key():
                return
            # make sure key is unique
            if key in parent_map:
                warn(f'Key {key} already exists.')
                # key = self.uniqueName(key, list(parent_map.keys()))
                return
            # swap keys in map
            parent_map[key] = parent_map.pop(self.key())
            # store new key in item
            self._data = key
        elif isinstance(parent_map, list):
            # Keys are not used for lists.
            # Instead, the sibling index is used as the list index.
            return
    
    def value(self) -> Any:
        if isinstance(self._data, dict) or isinstance(self._data, list):
            # this item stores the root key:value map
            return self._data
        
        # this item stores a key into its parent key:value map
        parent: KeyValueTreeItem = self.parent()
        if parent is None:
            return
        # get value from parent key:value mapping (dict or list)
        parent_map: dict | list = parent.value()
        if isinstance(parent_map, dict) or isinstance(parent_map, list):
            try:
                return parent_map[self.key()]
            except (KeyError, IndexError):
                return
    
    def setValue(self, value) -> None:
        if isinstance(self._data, dict) or isinstance(self._data, list):
            # this is the root item which must be a key:value map
            if not isinstance(value, dict) and not isinstance(value, list):
                return
            # reset the entire tree
            for child in self.children:
                child._parent = None
            self.children = []
            self._data = value
            self.setupSubtree()
            return
        
        # this item stores a key into its parent key:value map
        # remove any child items (in case this item was itself a key:value container)
        for child in self.children:
            child._parent = None
        self.children = []
        parent: KeyValueTreeItem = self.parent()
        if parent is None:
            self._data = value
        else:
            # set value in parent key:value mapping (dict or list)
            parent_map: dict | list = parent.value()
            parent_map[self.key()] = value
        self.setupSubtree()
    
    def setupSubtree(self):
        """ Recursively build subtree if value is itself a container with key:value access.
        """
        value = self.value()
        if isinstance(value, dict):
            for key in list(value.keys()):
                KeyValueTreeItem(key, parent=self)
        elif isinstance(value, list):
            for i in range(len(value)):
                # list keys are not explicitly set, they will default to the list index
                KeyValueTreeItem(None, parent=self)
    
    def name(self) -> str:
        key = self.key()
        return str(key) if key is not None else None
    
    def setName(self, name: str) -> None:
        self.setKey(name)
    
    def __repr__(self):
        key = self.key()
        value = self.value()
        if isinstance(value, dict):
            return f'{key}: {{}}'
        elif isinstance(value, list):
            return f'{key}: []'
        else:
            return f'{key}: {value}'
    
    def __str__(self) -> str:
        return self._tree_repr(lambda item: item.__repr__())
    
    def setParent(self, parent: KeyValueTreeItem | None) -> None:
        old_parent: KeyValueTreeItem | None = self.parent()
        new_parent: KeyValueTreeItem | None = parent
        if old_parent is new_parent:
            # nothing to do
            return
        if new_parent is not None:
            if new_parent.hasAncestor(self):
                raise ValueError('Cannot set parent to a descendant.')
            # the new parent must be a key:value container
            new_parent_value = new_parent.value()
            if not isinstance(new_parent_value, dict) and not isinstance(new_parent_value, list):
                raise ValueError('Parent must be a key:value mapping (dict or list).')
        
        value = self.value()
        
        if old_parent is not None:
            # detach from old parent
            old_parent_map: dict | list = old_parent.value()
            old_parent_map.pop(self.key())
            old_parent.children.remove(self)
        
        if new_parent is None:
            # root items must have a key:value mapping
            if isinstance(value, dict) or isinstance(value, list):
                self._data = value
            self._parent = None
        else:
            # attach to new parent (appends as last child)
            new_parent.children.append(self)
            self._parent = new_parent
            new_parent_map: dict | list = new_parent.value()
            # update new parent key:value mapping
            if isinstance(new_parent_map, dict):
                key = self.key()
                if (key is None) or (key == ''):
                    # ensure valid dict key
                    sibling_names = [sibling.name for sibling in self.siblings()]
                    key = self.uniqueName('?', sibling_names)
                if value is None:
                    value = self.value()
                # print('data:', self._data, 'key:', key, 'value:', value, 'parent dict:', new_parent_map)
                if (key not in new_parent_map) or (new_parent_map[key] is not value):
                    new_parent_map[key] = value
            elif isinstance(new_parent_map, list):
                if value is None:
                    value = self.value()
                # print('data:', self._data, 'value:', value, 'parent list:', new_parent_map)
                if len(new_parent_map) <= self.siblingIndex():
                    new_parent_map.append(value)
    
    def insertChild(self, index: int, child: KeyValueTreeItem) -> None:
        if not (0 <= index <= len(self.children)):
            raise IndexError('Index out of range.')
        # append as last child
        child.setParent(self)
        # move item to index
        pos = self.children.index(child)
        if pos != index:
            if pos < index:
                index -= 1
            if pos != index:
                this_map: dict | list = self.value()
                if isinstance(this_map, list):
                    this_map.insert(index, this_map.pop(pos))
                self.children.insert(index, self.children.pop(pos))


def test_tree():
    data = {
        'a': 1,
        'b': [4, 8, 9, 5, 7, 99],
        'c': {
            'me': 'hi',
            3: 67,
            'd': {
                'e': 3,
                'f': 'ya!',
                'g': 5,
            },
        },
    }
    root = KeyValueTreeItem(data)
    print('-'*82)
    print(root)
    import json
    print(json.dumps(data, indent='    '))

    print('-'*82)
    print('remove /a')
    root.removeChild(root['/a'])
    print(root)

    print('-'*82)
    print('remove /c/d/f')
    d = root['c/d']
    f = root['c/d/f']
    d.removeChild(f)
    print(root)

    print('-'*82)
    print('move /c/d to /d')
    d.setParent(root)
    print(root)

    print('-'*82)
    print('move /d to 2nd child of /')
    root.insertChild(1, d)
    print(root)

    print('-'*82)
    print('move /d to 3rd child of /b')
    b = root['b']
    b.insertChild(2, d)
    print(root)

    print('-'*82)
    print('remove /b/1')
    b.removeChild(b['1'])
    print(root)

    print('-'*82)
    print('/c/me:hi -> /c/me:bye')
    c = root['c']
    c.children[0].setValue('bye')
    print(root)
    print(c)

    print('-'*82)
    print('move /b/1 to first child of /c')
    c.insertChild(0, b['1'])
    print(root)


if __name__ == '__main__':
    test_tree()



""" PlotWidget with matlab color scheme and CustomPlotItem.
"""

from __future__ import annotations
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *
import platform
import pyqtgraph as pg
from pyqt_ext.pyqtgraph_ext import Plot


class Figure(pg.PlotWidget):
    """ PlotWidget with matlab color scheme and CustomPlotItem. """

    def __init__(self, *args, **kwargs):
        if 'plotItem' not in kwargs:
            kwargs['plotItem'] = Plot()
        pg.PlotWidget.__init__(self, *args, **kwargs)

        # MATLAB color scheme
        self.setBackground(QColor(240, 240, 240))

        if platform.system() == 'Darwin':
            # Fix error message due to touch events on MacOS trackpad.
            # !!! Warning: This may break touch events on a touch screen or mobile device.
            # See https://bugreports.qt.io/browse/QTBUG-103935
            for view in self.scene().views():
                view.viewport().setAttribute(Qt.WidgetAttribute.WA_AcceptTouchEvents, False)


def test_live():
    import numpy as np
    from pyqt_ext.pyqtgraph_ext import Graph
    app = QApplication()
    plot = Figure()
    view = plot.getViewBox()
    for i in range(3):
        line = Graph(y=np.random.randn(10), pen=pg.mkPen(color=view.nextColor(), width=3))
        plot.addItem(line)
    plot.setWindowTitle('pyqtgraph-tools')
    plot.show()
    # from pyqtgraph_ext import XAxisRegion
    # view.startDrawingItemsOfType(XAxisRegion)
    # QTimer.singleShot(2000, lambda: view.stopDrawingItems())
    app.exec()


if __name__ == '__main__':
    test_live()



from pyqt_ext.tree.AbstractTreeItem import AbstractTreeItem
from pyqt_ext.tree.AbstractTreeModel import AbstractTreeModel, AbstractTreeMimeData
from pyqt_ext.tree.TreeView import TreeView

from pyqt_ext.tree.KeyValueTreeItem import KeyValueTreeItem
from pyqt_ext.tree.KeyValueTreeModel import KeyValueTreeModel
from pyqt_ext.tree.KeyValueTreeView import KeyValueTreeView


from __future__ import annotations

"""Port of the widgets/layouts/flowlayout example from Qt v6.x"""

import sys
from qtpy.QtCore import Qt, QMargins, QPoint, QRect, QSize
from qtpy.QtWidgets import QLayout, QPushButton, QSizePolicy, QWidget


class FlowLayout(QLayout):
    def __init__(self, parent=None):
        super().__init__(parent)

        if parent is not None:
            self.setContentsMargins(QMargins(0, 0, 0, 0))

        self._item_list = []

    def __del__(self):
        item = self.takeAt(0)
        while item:
            item = self.takeAt(0)

    def addItem(self, item):
        self._item_list.append(item)

    def count(self):
        return len(self._item_list)

    def itemAt(self, index):
        if 0 <= index < len(self._item_list):
            return self._item_list[index]

        return None

    def takeAt(self, index):
        if 0 <= index < len(self._item_list):
            return self._item_list.pop(index)

        return None

    def expandingDirections(self):
        return Qt.Orientation(0)

    def hasHeightForWidth(self):
        return True

    def heightForWidth(self, width):
        height = self._do_layout(QRect(0, 0, width, 0), True)
        return height

    def setGeometry(self, rect):
        super(FlowLayout, self).setGeometry(rect)
        self._do_layout(rect, False)

    def sizeHint(self):
        return self.minimumSize()

    def minimumSize(self):
        size = QSize()

        for item in self._item_list:
            size = size.expandedTo(item.minimumSize())

        size += QSize(2 * self.contentsMargins().top(), 2 * self.contentsMargins().top())
        return size

    def _do_layout(self, rect, test_only):
        x = rect.x()
        y = rect.y()
        line_height = 0
        spacing = self.spacing()

        next_items = self._item_list[1:] + self._item_list[-1:]
        for item, next_item in zip(self._item_list, next_items):
            style: QStyle = item.widget().style()
            control_type = item.widget().sizePolicy().controlType()
            next_control_type = next_item.widget().sizePolicy().controlType()
            layout_spacing_x = style.layoutSpacing(control_type, next_control_type, Qt.Orientation.Horizontal)
            layout_spacing_y = style.layoutSpacing(control_type, next_control_type, Qt.Orientation.Vertical)
            space_x = spacing + layout_spacing_x
            space_y = spacing + layout_spacing_y
            next_x = x + item.sizeHint().width() + space_x
            if next_x - space_x > rect.right() and line_height > 0:
                x = rect.x()
                y = y + line_height + space_y
                next_x = x + item.sizeHint().width() + space_x
                line_height = 0

            if not test_only:
                item.setGeometry(QRect(QPoint(x, y), item.sizeHint()))

            x = next_x
            line_height = max(line_height, item.sizeHint().height())

        return y + line_height - rect.y()


if __name__ == "__main__":
    from qtpy.QtWidgets import *
    app = QApplication(sys.argv)
    ui = QWidget()
    flow = FlowLayout(ui)
    flow.setContentsMargins(0, 0, 0, 0)
    flow.setSpacing(0)
    flow.addWidget(QPushButton("Short"))
    flow.addWidget(QPushButton("Longer"))
    flow.addWidget(QPushButton("Different text"))
    flow.addWidget(QPushButton("More text"))
    flow.addWidget(QPushButton("Even longer button text"))
    flow.addWidget(QLabel("hi-ya!"))
    flow.addWidget(QLabel("text"))
    flow.addWidget(QCheckBox("cbox"))
    flow.addWidget(QRadioButton("radio"))
    flow.addWidget(QToolButton())
    flow.addWidget(QToolButton())
    flow.addWidget(QToolButton())
    flow.addWidget(QLineEdit())
    ui.show()
    sys.exit(app.exec())



""" Tree view for a `KeyValueTreeModel` with drag-and-drop, context menu, and mouse wheel expand/collapse.
"""

from __future__ import annotations
from typing import Callable
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *
from pyqt_ext.tree import TreeView, KeyValueTreeItem, KeyValueTreeModel


class KeyValueTreeView(TreeView):

    def __init__(self, *args, **kwargs) -> None:
        TreeView.__init__(self, *args, **kwargs)

        # delegate
        self.setItemDelegate(KeyValueTreeViewDelegate(self))
    
    def customContextMenu(self, index: QModelIndex = QModelIndex()) -> QMenu | None:
        model: KeyValueTreeModel = self.model()
        if model is None:
            return
        
        menu = QMenu(self)

        # context menu for item that was clicked on
        if index.isValid():
            item: KeyValueTreeItem = model.itemFromIndex(index)
            item_label = self.truncateLabel(item.path())
            menu.addAction(f'Insert New Item Before', lambda item=item: self.insertNewSiblingBefore(item))
            menu.addAction(f'Insert New Item After', lambda item=item: self.insertNewSiblingAfter(item))
            if not item.isLeaf():
                menu.addAction(f'Add New Child Item', lambda item=item: self.appendNewChild(item))
            menu.addSeparator()
            menu.addAction(f'Remove {item_label}', lambda item=item: self.askToRemoveItems([item]))
            menu.addSeparator()
        else:
            item: KeyValueTreeItem = model.root()
            menu.addAction(f'Add New Item', lambda item=item: self.appendNewChild(item))
            menu.addSeparator()
        
        self.appendDefaultContextMenu(menu)
        return menu
    
    def insertNewItem(self, parent_item: KeyValueTreeItem, row: int = -1) -> None:
        model: KeyValueTreeModel = self.model()
        if model is None:
            return
        
        parent_map: dict | list = parent_item.value()
        if isinstance(parent_map, dict):
            key = parent_item.uniqueName('New', list(parent_map.keys()))
        else:
            key = None
        if row == -1:
            row = len(parent_item.children)
        new_item = KeyValueTreeItem(key)
        model.insertItems(row, [new_item], parent_item)
    
    def insertNewSiblingBefore(self, item: KeyValueTreeItem) -> None:
        parent_item: KeyValueTreeItem = item.parent()
        row: int = item.siblingIndex()
        self.insertNewItem(parent_item, row)
    
    def insertNewSiblingAfter(self, item: KeyValueTreeItem) -> None:
        parent_item: KeyValueTreeItem = item.parent()
        row: int = item.siblingIndex() + 1
        self.insertNewItem(parent_item, row)
    
    def appendNewChild(self, parent_item: KeyValueTreeItem) -> None:
        row: int = -1
        self.insertNewItem(parent_item, row)


class KeyValueTreeViewDelegate(QStyledItemDelegate):
    """ Delegate for editing values.

    Provides checkbox for bool values.
    """
    def __init__(self, parent: QObject = None):
        QStyledItemDelegate.__init__(self, parent)
    
    def createEditor(self, parent: QWidget, option: QStyleOptionViewItem, index: QModelIndex):
        data = index.model().data(index, Qt.ItemDataRole.EditRole)
        if type(data) in [int, float, bool, str]:
            editor = QLineEdit(parent)
            editor.setText(str(data))
            return editor
        elif isinstance(data, tuple):
            editor = QLineEdit(parent)
            text = '(' + ', '.join(str(value) for value in data) + ')'
            editor.setText(text)
            return editor
        elif isinstance(data, list):
            editor = QLineEdit(parent)
            text = '[' + ', '.join(str(value) for value in data) + ']'
            editor.setText(text)
            return editor
        # elif isinstance(data, bool):
        #     # will handle with paint(), editorEvent(), and setModelData()
        #     return None
        return QStyledItemDelegate.createEditor(self, parent, option, index)

    def paint(self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex):
        data = index.model().data(index, Qt.ItemDataRole.DisplayRole)
        if type(data) in [tuple, list, bool]:
            if isinstance(data, tuple):
                text = '(' + ', '.join(str(value) for value in data) + ')'
            elif isinstance(data, list):
                text = '[' + ', '.join(str(value) for value in data) + ']'
            elif isinstance(data, bool):
                text = ' ' + str(data)
            if option.state & QStyle.StateFlag.State_Selected:
                painter.fillRect(option.rect, option.palette.highlight())
                painter.setPen(option.palette.highlightedText().color())
            else:
                painter.setPen(option.palette.text().color())
            painter.drawText(option.rect, Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter, text)
            return
        # elif isinstance(data, bool):
        #     # paint checkbox without label
        #     checked = data
        #     opts = QStyleOptionButton()
        #     opts.state |= QStyle.State_Active
        #     if index.flags() & Qt.ItemFlag.ItemIsEditable:
        #         opts.state |= QStyle.State_Enabled
        #     else:
        #         opts.state |= QStyle.State_ReadOnly
        #     if checked:
        #         opts.state |= QStyle.State_On
        #     else:
        #         opts.state |= QStyle.State_Off
        #     opts.rect = self.getCheckBoxRect(option)
        #     QApplication.style().drawControl(QStyle.CE_CheckBox, opts, painter)
        #     return
        return QStyledItemDelegate.paint(self, painter, option, index)

    def editorEvent(self, event: QEvent, model: KeyValueTreeModel, option: QStyleOptionViewItem, index: QModelIndex):
        # data = index.model().data(index, Qt.ItemDataRole.EditRole)
        # if isinstance(data, bool):
        #     # handle checkbox events
        #     if not (index.flags() & Qt.ItemFlag.ItemIsEditable):
        #         return False
        #     if event.button() == Qt.MouseButton.LeftButton:
        #         if event.type() == QEvent.MouseButtonRelease:
        #             if self.getCheckBoxRect(option).contains(event.pos()):
        #                 self.setModelData(None, model, index)
        #                 return True
        #         elif event.type() == QEvent.MouseButtonDblClick:
        #             if self.getCheckBoxRect(option).contains(event.pos()):
        #                 return True
        #     return False
        return QStyledItemDelegate.editorEvent(self, event, model, option, index)

    def setModelData(self, editor: QWidget, model: KeyValueTreeModel, index: QModelIndex):
        item: KeyValueTreeItem = model.itemFromIndex(index)
        old_value = item.value()
        if type(old_value) in [dict, list] and len(old_value) > 0:
            answer = QMessageBox.question(self.parent(), 'Overwrite?', 'Overwrite non-empty key:value map?', QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No, QMessageBox.StandardButton.No)
            if answer == QMessageBox.StandardButton.No:
                return
        if isinstance(editor, QLineEdit):
            new_value = str_to_value(editor.text())
            # did change involves a key:value map?
            mapping_changed = (type(old_value) in [dict, list]) or (type(new_value) in [dict, list])
            if mapping_changed:
                view: KeyValueTreeView = self.parent()
                view.storeState()
            model.setData(index, new_value, Qt.ItemDataRole.EditRole)
            if mapping_changed:
                view.restoreState()
            return
        # elif isinstance(data, bool):
        #     checked = not data
        #     model.setData(index, checked, Qt.ItemDataRole.EditRole)
        #     return
        return QStyledItemDelegate.setModelData(self, editor, model, index)

    def getCheckBoxRect(self, option: QStyleOptionViewItem):
        """ Get rect for checkbox positioned in option.rect.
        """
        # Get size of a standard checkbox
        opts = QStyleOptionButton()
        checkBoxRect = QApplication.style().subElementRect(QStyle.SE_CheckBoxIndicator, opts, None)
        # Position checkbox in option.rect
        x = option.rect.x()
        y = option.rect.y()
        w = option.rect.width()
        h = option.rect.height()
        # checkBoxTopLeftCorner = QPoint(x + w / 2 - checkBoxRect.width() / 2, y + h / 2 - checkBoxRect.height() / 2)  # horizontal center, vertical center
        checkBoxTopLeftCorner = QPoint(x, y + h / 2 - checkBoxRect.height() / 2)  # horizontal left, vertical center
        return QRect(checkBoxTopLeftCorner, checkBoxRect.size())


def str_to_value(text: str) -> bool | int | float | str | tuple | list | dict:
    if text.lower().strip() == 'true':
        return True
    if text.lower().strip() == 'false':
        return False
    if text.lstrip().startswith('(') and text.rstrip().endswith(')'):
        return tuple([str_to_value(item.strip()) for item in text.strip()[1:-1].split(',') if item.strip() != ''])
    if text.lstrip().startswith('[') and text.rstrip().endswith(']'):
        return [str_to_value(item.strip()) for item in text.strip()[1:-1].split(',') if item.strip() != '']
    if text.lstrip().startswith('{') and text.rstrip().endswith('}'):
        values = {}
        for field in [item for item in text.strip()[1:-1].split(',') if ':' in item]:
            key, value = field.split(':')
            values[key.strip()] = str_to_value(value.strip())
        return values
    try:
        return int(text)
    except ValueError:
        try:
            return float(text)
        except ValueError:
            return text


def test_live():
    data = {
        'a': 1,
        'b': [4, 8, (1, 5.5, True, 'good'), 5, 7, 99, True, False, 'hi', 'bye'],
        'c': {
            'me': 'hi',
            3: 67,
            'd': {
                'e': 3,
                'f': 'ya!',
                'g': 5,
            },
        },
    }

    app = QApplication()
    # root = KeyValueTreeItem(data)
    model = KeyValueTreeModel(data)
    view = KeyValueTreeView()
    view.setModel(model)
    view.show()
    view.resize(QSize(800, 600))
    view.showAll()
    # model.keyChanged.connect(lambda: print(model.root()))
    # model.valueChanged.connect(lambda: print(model.root()))
    app.exec()

    import json
    print(json.dumps(data, indent='    '))

if __name__ == '__main__':
    test_live()



""" PlotDataItem with custom context menu and style dialog.
"""

from __future__ import annotations
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *
from pyqt_ext.utils import toQColor
from pyqt_ext.widgets import TableWidgetWithCopyPaste
import pyqtgraph as pg
from pyqt_ext.pyqtgraph_ext import GraphStyle, editGraphStyle


class Graph(pg.PlotDataItem):
    """ PlotDataItem with custom context menu and style dialog. """

    sigNameChanged = Signal(str)

    def __init__(self, *args, **kwargs):
        # default style is first MATLAB line color
        if 'pen' not in kwargs:
            kwargs['pen'] = pg.mkPen(QColor(0, 114, 189), width=1)
        if 'symbolPen' not in kwargs:
            kwargs['symbolPen'] = pg.mkPen(QColor(0, 114, 189), width=1)
        if 'symbolBrush' not in kwargs:
            kwargs['symbolBrush'] = pg.mkBrush(QColor(0, 114, 189, 0))
        if 'symbol' not in kwargs:
            kwargs['symbol'] = None
        pg.PlotDataItem.__init__(self, *args, **kwargs)

        self.setZValue(1)

        self.contextMenu = QMenu()
        # self.contextMenu.addAction('Rename')
        # self.contextMenu.addSeparator()
        self.contextMenu.addAction('Data table', self.dataDialog)
        self.contextMenu.addSeparator()
        self.contextMenu.addAction('Style', self.styleDialog)
        # self.contextMenu.addSeparator()
        # self.contextMenu.addAction('Hide', lambda: self.setVisible(False))
        # self.contextMenu.addSeparator()
        # self.contextMenu.addAction('Delete', lambda: self.getViewBox().deleteItem(self))
    
    def hasCurve(self):
        pen = pg.mkPen(self.opts['pen'])
        return pen.style() != Qt.PenStyle.NoPen
    
    def hasSymbol(self):
        return 'symbol' in self.opts and self.opts['symbol'] is not None
    
    def shape(self) -> QPainterPath:
        if self.hasCurve():
            return self.curve.shape()
        elif self.hasSymbol():
            return self.scatter.shape()

    def boundingRect(self):
        return self.shape().boundingRect()
    
    def mouseClickEvent(self, event):
        if event.button() == Qt.RightButton:
            if self.hasCurve():
                if self.curve.mouseShape().contains(event.pos()):
                    if self.raiseContextMenu(event):
                        event.accept()
                        return
            if self.hasSymbol():
                if len(self.scatter.pointsAt(event.pos())) > 0:
                    if self.raiseContextMenu(event):
                        event.accept()
                        return
    
    def raiseContextMenu(self, event):
        menu = self.getContextMenus(event)
        pos = event.screenPos()
        menu.popup(QPoint(int(pos.x()), int(pos.y())))
        return True
    
    def getContextMenus(self, event=None):
        name = self.name()
        if name is None:
            name = self.__class__.__name__
        self.contextMenu.setTitle(name)

        self.menu = QMenu()
        self.menu.addMenu(self.contextMenu)

        # Let the scene add on to the end of our context menu (this is optional)
        self.menu.addSection('View')
        self.menu = self.scene().addParentContextMenus(self, self.menu, event)
        return self.menu
    
    def name(self):
        return self.opts.get('Name', None)
    
    def setName(self, name):
        if name is None:
            del self.opts['Name']
        else:
            self.opts['Name'] = name
        self.sigNameChanged.emit(self.name())
    
    def graphStyle(self) -> GraphStyle:
        style = GraphStyle()

        pen = pg.mkPen(self.opts['pen'])
        symbolPen = pg.mkPen(self.opts['symbolPen'])
        symbolBrush = pg.mkBrush(self.opts['symbolBrush'])

        style['color'] = pen.color()
        style['linestyle'] = pen.style()
        style['linewidth'] = pen.widthF()
        style['marker'] = self.opts.get('symbol', None)
        style['markersize'] = self.opts.get('symbolSize', 10)
        style['markeredgestyle'] = symbolPen.style()
        style['markeredgewidth'] = symbolPen.widthF()
        if symbolPen.color() != pen.color():
            style['markeredgecolor'] = symbolPen.color()
        if symbolBrush.color() != symbolPen.color():
            style['markerfacecolor'] = symbolBrush.color()

        return style
    
    def setGraphStyle(self, style: GraphStyle, colorIndex: int | None = None) -> int | None:
        # color
        color = style['color']
        if color is None:
            if colorIndex is not None:
                try:
                    axes = self.getViewBox()
                    colormap = axes.colormap()
                    color = colormap[colorIndex % len(colormap)]
                    color = toQColor(color)
                    colorIndex += 1
                except:
                    oldStyle = self.graphStyle()
                    color = toQColor(oldStyle['color'])
            else:
                oldStyle = self.graphStyle()
                color = toQColor(oldStyle['color'])
        else:
            color = toQColor(color)

        # line
        lineStyle: str = style['linestyle']
        linePenStyle: Qt.PenStyle = GraphStyle.penStyles[GraphStyle.lineStyles.index(lineStyle)]
        lineWidth = style['linewidth']
        linePen = pg.mkPen(color=color, width=lineWidth, style=linePenStyle)
        self.setPen(linePen)

        # marker
        marker = style['marker']
        if isinstance(marker, str) and marker.lower() == 'none':
            marker = None
        marker_labels = list(GraphStyle.pyqtgraphMarkers.keys())
        marker_keys = list(GraphStyle.pyqtgraphMarkers.values())
        if marker in marker_labels:
            marker = marker_keys[marker_labels.index(marker)]
        self.setSymbol(marker)
        
        markerSize = style['markersize']
        self.setSymbolSize(markerSize)

        markerEdgeStyle = style['markeredgestyle']
        markerEdgePenStyle: Qt.PenStyle = GraphStyle.penStyles[GraphStyle.lineStyles.index(markerEdgeStyle)]
        markerEdgeWidth = style['markeredgewidth']
        markerEdgeColor = style['markeredgecolor']
        if markerEdgeColor is None:
            markerEdgeColor = color
        else:
            markerEdgeColor = toQColor(markerEdgeColor)
        symbolPen = pg.mkPen(color=markerEdgeColor, width=markerEdgeWidth, style=markerEdgePenStyle)
        self.setSymbolPen(symbolPen)

        markerFaceColor = style['markerfacecolor']
        if markerFaceColor is None:
            markerFaceColor = markerEdgeColor
        else:
            markerFaceColor = toQColor(markerFaceColor)
        self.setSymbolBrush(markerFaceColor)
        
        return colorIndex
    
    def styleDialog(self):
        name = self.name()
        if name is None:
            name = self.__class__.__name__
        old_style: GraphStyle = self.graphStyle()
        new_style: GraphStyle | None = editGraphStyle(old_style, parent = self.getViewBox().getViewWidget(), title = name)
        if new_style is None:
            return
        self.setGraphStyle(new_style)
    
    def dataDialog(self):
        dlg = QDialog()
        dlg.setWindowTitle(self.name())
        vbox = QVBoxLayout(dlg)
        vbox.setContentsMargins(0, 0, 0, 0)
        vbox.setSpacing(0)
        
        xdata, ydata = self.getOriginalDataset()
        n_rows = len(ydata)
        n_cols = 2
        table = TableWidgetWithCopyPaste(n_rows, n_cols)
        for row in range(n_rows):
            table.setItem(row, 0, QTableWidgetItem(str(xdata[row])))
            table.setItem(row, 1, QTableWidgetItem(str(ydata[row])))
        table.setEditTriggers(QAbstractItemView.NoEditTriggers)

        xlabel = getattr(self, '_xlabel', None)
        xunits = getattr(self, '_xunits', None)
        ylabel = getattr(self, '_ylabel', None)
        yunits = getattr(self, '_yunits', None)
        
        view: pg.ViewBox = self.getViewBox()
        plot: pg.PlotItem = view.parentWidget()

        if (xlabel is None) or (xunits is None):
            # try and get label and units from plot axis
            xaxis = plot.getAxis('bottom')
            if xlabel is None:
                xlabel = xaxis.labelText
            if not xlabel:
                # check linked views for valid axis label
                linked_views = view.getState()['linkedViews'][0]
                if linked_views is not None:
                    for linked_view in linked_views:
                        linked_plot: pg.PlotItem = linked_view.parentWidget()
                        linked_xaxis = linked_plot.getAxis('bottom')
                        if linked_xaxis.labelText:
                            xaxis = linked_xaxis
                            xlabel = xaxis.labelText
                            break
            if xunits is None:
                xunits = xaxis.labelUnits
        if xunits:
            xlabel += f' ({xunits})'
        
        if (ylabel is None) or (yunits is None):
            # try and get label and units from plot axis
            yaxis = plot.getAxis('left')
            if ylabel is None:
                ylabel = yaxis.labelText
            if not ylabel:
                # check linked views for valid axis label
                linked_views = view.getState()['linkedViews'][1]
                if linked_views is not None:
                    for linked_view in linked_views:
                        linked_plot: pg.PlotItem = linked_view.parentWidget()
                        linked_yaxis = linked_plot.getAxis('left')
                        if linked_yaxis.labelText:
                            yaxis = linked_yaxis
                            ylabel = yaxis.labelText
                            break
            if yunits is None:
                yunits = yaxis.labelUnits
        if yunits:
            ylabel += f' ({yunits})'
        
        table.setHorizontalHeaderLabels([xlabel, ylabel])
        for col in range(n_cols):
            table.resizeColumnToContents(col)
        table.resizeRowsToContents()
        vbox.addWidget(table)
        dlg.exec()



""" PySide/PyQt expandable/collapsible section.
"""

from qtpy.QtCore import Qt, QPropertyAnimation, QParallelAnimationGroup, QAbstractAnimation
from qtpy.QtWidgets import QWidget, QToolButton, QFrame, QScrollArea, QPushButton, QLineEdit, QSizePolicy, QLayout, QFormLayout, QVBoxLayout, QGridLayout
try:
    import qtawesome as qta
except ImportError:
    qta = None


class CollapsibleSection(QWidget):
    def __init__(self, parent: QWidget = None, title: str = '', animationDuration: int = 250) -> None:
        """
        References:
            https://stackoverflow.com/questions/32476006/how-to-make-an-expandable-collapsable-section-widget-in-qt
            # Adapted from PyQt4 version
            https://stackoverflow.com/a/37927256/386398
            # Adapted from c++ version
            https://stackoverflow.com/a/37119983/386398
        """
        QWidget.__init__(self, parent=parent)

        # toggle button
        self.toggleButton = QToolButton()
        self.toggleButton.setStyleSheet("QToolButton { border: none; }")
        self.toggleButton.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
        if qta:
            # Use font awesome icons because default arrow icons on MacOS look terrible
            self.toggleButton.setIcon(qta.icon('fa6s.angle-right'))
        else:
            self.toggleButton.setArrowType(Qt.RightArrow)
        self.toggleButton.setText(str(title))
        self.toggleButton.setCheckable(True)
        self.toggleButton.setChecked(False)

        # header line
        self.headerLine = QFrame()
        self.headerLine.setFrameShape(QFrame.HLine)
        self.headerLine.setFrameShadow(QFrame.Sunken)
        self.headerLine.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Maximum)

        # content area
        self.contentArea = QScrollArea()
        self.contentArea.setStyleSheet("QScrollArea { border: none; }")
        self.contentArea.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        # start out collapsed
        self.contentArea.setMaximumHeight(0)
        self.contentArea.setMinimumHeight(0)
        
        # layout
        self.mainLayout = QGridLayout()
        self.mainLayout.setVerticalSpacing(0)
        self.mainLayout.setContentsMargins(0, 0, 0, 0)
        row = 0
        self.mainLayout.addWidget(self.toggleButton, row, 0, 1, 1, Qt.AlignmentFlag.AlignLeft)
        self.mainLayout.addWidget(self.headerLine, row, 2, 1, 1)
        row = 1
        self.mainLayout.addWidget(self.contentArea, row, 0, 1, 3)
        self.setLayout(self.mainLayout)
        
        # toggle animation
        self.animationDuration = animationDuration
        self.toggleAnimation = QParallelAnimationGroup()
        self.toggleAnimation.addAnimation(QPropertyAnimation(self, b"minimumHeight"))
        self.toggleAnimation.addAnimation(QPropertyAnimation(self, b"maximumHeight"))
        self.toggleAnimation.addAnimation(QPropertyAnimation(self.contentArea, b"maximumHeight"))

        self.toggleButton.clicked.connect(self.setIsExpanded)

    def setTitle(self, title: str) -> None:
        self.toggleButton.setText(title)
    
    def setContentLayout(self, contentLayout: QLayout) -> None:
        self.contentArea.destroy()
        self.contentArea.setLayout(contentLayout)

        # update animations
        collapsedHeight = self.sizeHint().height() - self.contentArea.maximumHeight()
        contentHeight = contentLayout.sizeHint().height()

        minimumHeightAnimation = self.toggleAnimation.animationAt(0)
        maximumHeightAnimation = self.toggleAnimation.animationAt(1)
        for anim in [minimumHeightAnimation, maximumHeightAnimation]:
            anim.setDuration(self.animationDuration)
            anim.setStartValue(collapsedHeight)
            anim.setEndValue(collapsedHeight + contentHeight)

        contentAreaMaximumHeightAnimation = self.toggleAnimation.animationAt(2)
        contentAreaMaximumHeightAnimation.setDuration(self.animationDuration)
        contentAreaMaximumHeightAnimation.setStartValue(0)
        contentAreaMaximumHeightAnimation.setEndValue(contentHeight)
    
    def isExpanded(self) -> bool:
        return self.toggleButton.isChecked()
    
    def setIsExpanded(self, expanded: bool) -> None:
        if self.toggleButton.isChecked() != expanded:
            self.toggleButton.setChecked(expanded)
        if qta:
            icon = qta.icon('fa6s.angle-down') if expanded else qta.icon('fa6s.angle-right')
            self.toggleButton.setIcon(icon)
        else:
            arrow_type = Qt.DownArrow if expanded else Qt.RightArrow
            self.toggleButton.setArrowType(arrow_type)
        direction = QAbstractAnimation.Direction.Forward if expanded else QAbstractAnimation.Direction.Backward
        self.toggleAnimation.setDirection(direction)
        self.toggleAnimation.start()
    
    def expand(self) -> None:
        self.setIsExpanded(True)
    
    def collapse(self) -> None:
        self.setIsExpanded(False)


def test_live():
    from qtpy.QtCore import QTimer
    from qtpy.QtWidgets import QApplication

    app = QApplication()

    ui = QWidget()
    vbox = QVBoxLayout(ui)

    details = CollapsibleSection(title='Details')

    form = QFormLayout()
    form.addRow('Red', QPushButton('Red'))
    form.addRow('Name', QLineEdit('Bob'))

    details.setTitle('Stuff')
    details.setContentLayout(form)

    vbox.addWidget(details)
    ui.show()

    QTimer.singleShot(1000, lambda: details.expand())
    QTimer.singleShot(2000, lambda: details.collapse())

    app.exec()


if __name__ == '__main__':
    test_live()



""" PlotWidget with matlab color scheme and CustomPlotItem.
"""

from __future__ import annotations
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *
import numpy as np
import platform
import pyqtgraph as pg
from pyqt_ext.pyqtgraph_ext import Plot


class PlotGrid(pg.GraphicsLayoutWidget):
    """ Grid of PlotItems. """

    def __init__(self, rows=0, cols=0, *args, **kwargs):
        pg.GraphicsLayoutWidget.__init__(self, *args, **kwargs)

        self._graphics_layout: pg.GraphicsLayout = self.ci

        self._grid_layout: QGraphicsGridLayout = self.ci.layout
        self._grid_layout.setContentsMargins(0, 0, 0, 0)
        self._grid_layout.setSpacing(0)

        # MATLAB color scheme
        self.setBackground(QColor(240, 240, 240))

        if platform.system() == 'Darwin':
            # Fix error message due to touch events on MacOS trackpad.
            # !!! Warning: This may break touch events on a touch screen or mobile device.
            # See https://bugreports.qt.io/browse/QTBUG-103935
            for view in self.scene().views():
                view.viewport().setAttribute(Qt.WidgetAttribute.WA_AcceptTouchEvents, False)
        
        if rows * cols > 0:
            self.setGrid(rows, cols)
    
    def rowCount(self) -> int:
        return self._grid_layout.rowCount()
            
    def columnCount(self) -> int:
        return self._grid_layout.columnCount()
    
    def clear(self) -> None:
        for item in list(self.items()):
            self.removeItem(item)
    
    def setGrid(self, rows: int, cols: int, plotType = Plot) -> None:
        for row in range(rows):
            for col in range(cols):
                item = self.getItem(row, col)
                if not issubclass(type(item), pg.PlotItem):
                    if item:
                        self.removeItem(item)
                    plot = plotType()
                    self.addItem(plot, row, col)
        for row in reversed(range(rows, self.rowCount())):
            for col in range(self.columnCount()):
                item = self.getItem(row, col)
                if item:
                    self.removeItem(item)
        for col in reversed(range(cols, self.columnCount())):
            for row in range(self.rowCount()):
                item = self.getItem(row, col)
                if item:
                    self.removeItem(item)
        if self.hasRegularLayout():
            self.applyRegularLayout()
    
    def plots(self) -> list[pg.PlotItem]:
        return [item for item in self.items() if issubclass(type(item), pg.PlotItem)]
    
    def hasRegularLayout(self) -> bool:
        return getattr(self, '_hasRegularLayout', False)
    
    def setHasRegularLayout(self, value: bool) -> None:
        self._hasRegularLayout = value
        if value:
            self.applyRegularLayout()
    
    def applyRegularLayout(self) -> None:
        if self.rowCount() * self.columnCount() <= 1:
            return
        viewWidth = 0
        n = 0
        for col in range(self.columnCount()):
            item = self.getItem(0, col)
            if issubclass(type(item), pg.PlotItem):
                viewWidth += item.getViewBox().width()
                n += 1
        viewWidth /= n
        viewWidth = int(viewWidth)

        viewHeight = 0
        n = 0
        for row in range(self.rowCount()):
            item = self.getItem(row, 0)
            if issubclass(type(item), pg.PlotItem):
                viewHeight += item.getViewBox().height()
                n += 1
        viewHeight /= n
        viewHeight = int(viewHeight)

        for row in range(self.rowCount()):
            for col in range(self.columnCount()):
                plot = self.getItem(row, col)
                if issubclass(type(plot), pg.PlotItem):
                    xaxis = plot.getAxis('bottom')
                    yaxis = plot.getAxis('left')
                    plot.setPreferredWidth(viewWidth + yaxis.width() if yaxis.isVisible() else viewWidth)
                    plot.setPreferredHeight(viewHeight + xaxis.height() if xaxis.isVisible() else viewHeight)
    
    def setAxisLabelAndTickVisibility(self, 
        xlabel_rows: list[int] = None,
        xtick_rows: list[int] = None,
        ylabel_columns: list[int] = None,
        ytick_columns: list[int] = None,
    ) -> None:
        # this accounts for any negative indexing
        rows = list(range(self.rowCount()))
        columns = list(range(self.columnCount()))
        xlabel_rows = rows if xlabel_rows is None else [rows[row] for row in xlabel_rows]
        xtick_rows = rows if xtick_rows is None else [rows[row] for row in xtick_rows]
        ylabel_columns = columns if ylabel_columns is None else [columns[col] for col in ylabel_columns]
        ytick_columns = columns if ytick_columns is None else [columns[col] for col in ytick_columns]
        # update axes
        for row in range(self.rowCount()):
            for col in range(self.columnCount()):
                plot = self.getItem(row, col)
                if not issubclass(type(plot), pg.PlotItem):
                    continue
                xaxis = plot.getAxis('bottom')
                yaxis = plot.getAxis('left')
                if row in xlabel_rows:
                    xaxis.label.show()
                else:
                    xaxis.label.hide()
                if col in ylabel_columns:
                    yaxis.label.show()
                else:
                    yaxis.label.hide()
                xaxis.setStyle(showValues=(row in xtick_rows))
                yaxis.setStyle(showValues=(col in ytick_columns))
    
    def resizeEvent(self, event: QResizeEvent) -> None:
        super().resizeEvent(event)
        if self.hasRegularLayout():
            self.applyRegularLayout()

def test_live():
    app = QApplication()
    grid = PlotGrid(3, 4)
    grid.setAxisLabelAndTickVisibility(xlabel_rows=[-1], xtick_rows=[-1], ylabel_columns=[0], ytick_columns=[0])
    grid.setHasRegularLayout(True)
    grid.setWindowTitle('pyqtgraph-tools.PlotGrid')
    grid.show()
    QTimer.singleShot(1000, lambda: grid.applyRegularLayout())
    app.exec()


if __name__ == '__main__':
    test_live()



from pyqt_ext.pyqtgraph_ext.GraphStyle import GraphStyle, GraphStylePanel, editGraphStyle
from pyqt_ext.pyqtgraph_ext.Graph import Graph

from pyqt_ext.pyqtgraph_ext.AxisRegion import AxisRegion, XAxisRegion, YAxisRegion, editAxisRegion, formatAxisRegion

from pyqt_ext.pyqtgraph_ext.View import View
from pyqt_ext.pyqtgraph_ext.Plot import Plot
from pyqt_ext.pyqtgraph_ext.Figure import Figure
from pyqt_ext.pyqtgraph_ext.PlotGrid import PlotGrid
